

#[
from __future__ import annotations

from . import solutions as _solutions
#]


def run_forward(
    solution: _solutions.Solution,
    initials: tuple[_np.ndarray, _np.ndarray, ],
    /,
) -> None:
    """
    """
    a0, P0 = initials
    Ta = solution.Ta
    Ka = solution.Ka
    Pa = solution.Pa




"""
"""


#[
from __future__ import annotations

from .. import quantities as _quantities
from .. import pages as _pages
from ..databoxes import main as _databoxes
#]


class AssignInlay:
    """
    """
    #[

    @_pages.reference(category="parameters", )
    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
················································································

==Assign model parameters==

```
self.assign(
    name_one=value_one,
    name_two=value_two,
    # etc
)
```

```
self.assign(databox, )
```

...


### Input arguments ###


???+ input "self"

    `Sequential` model whose parameters will be assigned.


???+ input "name_one"

    Name of a parameter to assign.


???+ input "value_one"

    Value to assign to `name_one`.

etc...

???+ input "databox"

    `Databox` or `dict` from which the parameters will be extracted and
    assigned.

················································································
        """
        for v in self._variants:
            v.assign(*args, **kwargs, )

    def get_parameters(self, /, ) -> _databoxes.Databox:
        """
        """
        parameter_names = self._invariant.parameter_names
        if self.is_singleton:
            return _databoxes.Databox(**{
                n: self._variants[0].parameters[n]
                for n in parameter_names
            })
        else:
            return _databoxes.Databox(**{
                n: [ v.parameters[n] for v in self._variants ]
                for n in parameter_names
            })

    #]






import sys
sys.path.append("../../src")
import irispie as _ip
import numpy as _np


x = _ip.Series.from_start_date_and_data(
    _ip.qq(2020, 1),
    range(10),
)

y = _ip.Series.from_start_date_and_data(
    x.start_date-5, _np.random.standard_normal((20,1))
)

fx = _ip.cum_diff(x, initial=y, )
fx2 = _ip.cum_diff(x, -2, initial=y, )

bx2  = _ip.cum_diff(x, initial=y, range=_ip.Ranger(None,None,-1))

w0 = _ip.Series.from_start_date_and_data(
    _ip.qq(2018,1),
    _np.random.standard_normal((36,1)),
)

dw0 =  _ip.diff(w0)

w1 = _ip.cum_diff(dw0, initial=0, range=_ip.Ranger(step=-1,))

dw = _ip.Series.from_start_date_and_data(
    _ip.qq(2020,1),
    _np.random.standard_normal((12,1)),
)

w2 = _ip.cum_diff(dw, "soy", initial=0, range=_ip.qq(2021,1)>>_ip.end)
w1 = _ip.cum_diff(dw, "eopy", initial=dw, range=_ip.qq(2021,1)>>_ip.end)




"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, )
import numpy as _np

from ..dataslates import main as _dataslates
from ..plans import main as _plans
from ..fords import solutions as _solutions
from ..fords import descriptors as _descriptors
#]


def simulate_flat(
    solution: _solutions.Solution,
    solution_vectors: _descriptors.SolutionVectors,
    dataslate: _dataslates.HorizontalDataslate,
    vid: int,
    /,
    *,
    plan: _plans.PlanSimulate | None = None,
    deviation: bool = False,
) -> None:
    """
    """
    #[
    columns_to_run = dataslate.base_columns
    boolex_logly = dataslate.boolex_logly
    working_data = dataslate.get_data_variant(0, )
    #
    column_start = columns_to_run[0]
    column_slice = slice(column_start, column_start+len(columns_to_run))
    column_array = _np.array(range(column_start, column_start+len(columns_to_run)))
    #
    vec = solution_vectors
    #
    T = solution.T
    P = solution.P
    R = solution.R
    K = solution.K if not deviation else 0
    #
    Z = solution.Z
    H = solution.H
    D = solution.D if not deviation else 0
    #
    if any(boolex_logly):
        working_data[boolex_logly, :] = _np.log(working_data[boolex_logly, :])
    #
    curr_state = _dataslates.retrieve_vector_from_data_array(
        working_data, vec.transition_variables, column_start-1,
    )
    #
    missing_initials = _np.isnan(curr_state, )
    unnecessary_initials = ~_np.array(vec.are_initial_conditions, )
    missing_unnecessary_initials = missing_initials & unnecessary_initials
    curr_state[missing_unnecessary_initials] = 0
    #
    no_shift_state_to_slab_lhs = [
        t.qid
        for t in vec.transition_variables
        if t.shift == 0
    ]
    #
    no_shift_state_to_slab_rhs = [
        j 
        for j, t in enumerate(vec.transition_variables)
        if t.shift == 0
    ]
    #
    measurement_variables_index = [t.qid for t in vec.measurement_variables]
    #
    unanticipated_shocks = _extract_shock_values(working_data, vec.unanticipated_shocks, )
    anticipated_shocks = _extract_shock_values(working_data, vec.anticipated_shocks, )
    measurement_shocks = _extract_shock_values(working_data, vec.measurement_shocks, )
    #
    Rx = [R]
    forward = None
    #
    anticipated_shock_column_incidence = list(_np.any(anticipated_shocks[:, column_array] != 0, axis=0))
    if any(anticipated_shock_column_incidence):
        # Last shock at t+forward
        forward = len(anticipated_shock_column_incidence) - anticipated_shock_column_incidence[::-1].index(True) - 1
        Rk = solution.expand_square_solution(forward, )
        Rx = Rx + (Rk if Rk is not None else [])
    #
    shock_column_end = (
        column_start + forward
        if forward is not None
        else column_start - 1
    )
    #
    for t in column_array:
        anticipated_shock_impact = sum(
            Rx[k] @ anticipated_shocks[:, s] if Rx[k] is not None else 0
            for k, s in enumerate(range(t, shock_column_end+1))
        )
        curr_state = T @ curr_state + P @ unanticipated_shocks[:, t] + anticipated_shock_impact + K
        working_data[no_shift_state_to_slab_lhs, t] = curr_state[no_shift_state_to_slab_rhs]
        #
        y = Z @ curr_state + H @ measurement_shocks[:, t] + D
        working_data[measurement_variables_index, t] = y
        #
    _store_shock_values(unanticipated_shocks, working_data, vec.unanticipated_shocks, )
    _store_shock_values(anticipated_shocks, working_data, vec.anticipated_shocks, )
    _store_shock_values(measurement_shocks, working_data, vec.measurement_shocks, )
    #
    if any(boolex_logly):
        working_data[boolex_logly, :] = _np.exp(working_data[boolex_logly, :])
    #]


def _extract_shock_values(working_data, shock_vector, ) -> _np.ndarray:
    """
    """
    #[
    shock_qids = [t.qid for t in shock_vector]
    shock_values = working_data[shock_qids, :]
    _np.nan_to_num(shock_values, copy=False, nan=0.0, posinf=0.0, neginf=0.0, )
    return shock_values
    #]


def _store_shock_values(shock_values, working_data, shock_vector, ) -> None:
    """
    """
    #[
    shock_qids = [t.qid for t in shock_vector]
    working_data[shock_qids, :] = shock_values
    #]




"""
Univariate time series filters
"""


#[
from __future__ import annotations

from numbers import (Real, )
from collections.abc import (Iterable, Callable, )
from types import (EllipsisType, )
from typing import (Self, )
import numpy as _np

from .. import dates as _dates
from . import main as _series
from . import _functionalize
#]


__all__ = ("hpf", )


_AUTO_SMOOTH = {
    _dates.Frequency.YEARLY: (10*1)**2,
    _dates.Frequency.HALFYEARLY: (10*2)**2,
    _dates.Frequency.QUARTERLY: (10*4)**2,
    _dates.Frequency.MONTHLY: (10*12)**2,
    "default": (10*4)**2,
}


def _get_default_smooth(frequency, /, ):
    """
    Default smoothing parameter (lambda)
    """
    return _AUTO_SMOOTH.get(frequency, _AUTO_SMOOTH["default"])


class _ConstrainedHodrickPrescottFilter:
    """
    """
    #[
    def __init__(
        self,
        num_periods: int,
        smooth: Real,
        /,
        level_where: list[int] | None = None,
        change_where: list[int] | None = None,
        log: bool = False,
    ) -> None:
        self._num_periods = num_periods
        self._smooth = smooth
        self._log = log
        self._num_extra_rows = 0
        self._create_plain_filter_matrix()
        self._add_level_constraints(level_where, )
        self._add_change_constraints(change_where, )

    def filter_data(
        self,
        data: _np.ndarray,
        /,
        level_data: _np.ndarray | None = None,
        change_data: _np.ndarray | None = None,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        """
        data = data.reshape(-1, 1)
        F = self._add_eye_for_observations(data, )
        extended_data = self._extend_data(data, level_data, change_data, )
        if self._log:
            extended_data = _np.log(extended_data, )
        enforced_zeros_where = _np.where(_np.isnan(data, ))
        extended_data[enforced_zeros_where] = 0
        trend_data = _np.linalg.solve(F, extended_data, )
        extended_data[enforced_zeros_where] = _np.nan
        gap_data = extended_data - trend_data
        if self._num_extra_rows > 0:
            trend_data = trend_data[:-self._num_extra_rows, :]
            gap_data = gap_data[:-self._num_extra_rows, :]
        if self._log:
            trend_data = _np.exp(trend_data, )
            gap_data = _np.exp(gap_data, )
        return trend_data, gap_data

    def _add_eye_for_observations(
        self,
        data: _np.ndarray,
        /,
    ) -> None:
        """
        Add 1 to the F diagonal for each row where an observation is available
        """
        #[
        F = _np.copy(self._F)
        quasi_eye = _np.diag(_np.float64(~_np.isnan(data.flatten(), )))
        F[:self._num_periods, :self._num_periods] += quasi_eye
        return F
        #]

    def _extend_data(
        self,
        data: _np.ndarray,
        level_data: _np.ndarray | None,
        change_data: _np.ndarray | None,
    ) -> _np.ndarray:
        """
        """
        extended_data = data.copy() if data.ndim > 1 else data.copy().reshape(-1, 1)
        tile_dim = (1, extended_data.shape[1], )
        if level_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(level_data, tile_dim), ), )
        if change_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(change_data, tile_dim), ), )
        return extended_data

    def _create_plain_filter_matrix(self, ) -> None:
        """
        """
        K = _np.zeros((self._num_periods-2, self._num_periods), dtype=float)
        for i in range(self._num_periods-2):
            K[i,i] = 1
            K[i,i+2] = 1
        for i in range(self._num_periods-2):
            K[i,i+1] = -2
        self._F = self._smooth * (K.T @ K)

    def _add_level_constraints(self, level_where: list[int], /, ):
        if not level_where:
            return
        num_constraints = len(level_where)
        self._num_extra_rows += num_constraints
        extra_rows = _np.zeros((num_constraints, self._num_periods, ), dtype=float)
        extra_variants = _np.zeros((self._num_periods + num_constraints, num_constraints, ), dtype=float)
        for i, j in enumerate(level_where, ):
            extra_rows[i, j] = 1
            extra_variants[j, i] = 1
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_variants, ))

    def _add_change_constraints(self, change_where: list[int], /, ):
        if not change_where:
            return
        num_constraints = len(change_where)
        extra_rows = _np.zeros((num_constraints, self._F.shape[1], ), dtype=float, )
        extra_variants = _np.zeros((self._F.shape[0] + num_constraints, num_constraints, ), dtype=float, )
        for i, j in enumerate(change_where, ):
            extra_rows[i, [j-1, j]] = (-1, 1)
            extra_variants[[j-1, j], i] = (-1, 1)
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_variants, ))
        self._num_extra_rows += num_constraints
    #]


class Mixin:
    """
    Constrained Hodrick-Prescott filter mixin for Series
    """
    #[

    def hpf_trend(self, /, *args, **kwargs):
        start_date, trend_data, _ = _data_hpf(self, *args, **kwargs, )
        self._replace_start_date_and_values(start_date, trend_data, )

    def hpf_gap(self, /, *args, **kwargs):
        start_date, _, gap_data = _data_hpf(self, *args, **kwargs, )
        self._replace_start_date_and_values(start_date, gap_data, )

    #]


def _data_hpf(
    self,
    *,
    span: Iterable[_dates.Dater] | EllipsisType = ...,
    smooth: Real | None = None,
    log: bool = False,
    level: _series.Series | None = None,
    change: _series.Series | None = None,
) -> tuple[_np.ndarray, _np.ndarray, ]:
    """
    Hodrick-Prescott filter run on a multi-variant data matrix
    """
    #[
    if smooth is None:
        smooth = _get_default_smooth(self.frequency, )
    span = self._resolve_dates(span, )
    encompassing_span, *from_to = _dates.get_encompassing_span(self, level, change, span, )
    num_periods = len(encompassing_span, )
    level_data, level_where = _prepare_constraints(level, from_to, )
    change_data, change_where = _prepare_constraints(change, from_to, )
    change_data, change_where = _remove_first_date_change(change_data, change_where, )
    #
    hp = _ConstrainedHodrickPrescottFilter(
        num_periods,
        smooth,
        level_where=level_where,
        change_where=change_where,
        log=log,
    )
    #
    trend_data = []
    gap_data = []
    for data_variant in self.iter_own_data_variants_from_to(from_to, ):
        trend_data_variant, gap_data_variant = hp.filter_data(
            data_variant,
            level_data=level_data,
            change_data=change_data,
        )
        trend_data.append(trend_data_variant, )
        gap_data.append(gap_data_variant, )
    trend_data = _np.hstack(trend_data, )
    gap_data = _np.hstack(gap_data, )
    #
    new_start_date = from_to[0]
    return new_start_date, trend_data, gap_data
    #]


def hpf(self, *args, **kwargs, ) -> tuple[_series.Series, _series.Series]:
    """
    Constrained Hodrick-Prescott filter
    """
    start_date, trend_data, gap_data = _data_hpf(self, *args, **kwargs, )
    trend = type(self)(start_date=start_date, values=trend_data, )
    gap = type(self)(start_date=start_date, values=gap_data, )
    return trend, gap


for n in ("hpf_trend", "hpf_gap", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )



def _prepare_constraints(
    constraint: Self | None,
    from_to: tuple[_dates.Dater, _dates.Dater, ],
    /,
) -> tuple[_np.ndarray | None, list[int] | None, ]:
    #[
    if constraint is None:
        return None, None
    data = constraint.get_data_from_to(from_to, 0, )
    where = list(_np.where(~_np.isnan(data.reshape(-1, ), ), )[0])
    data = data[where, :]
    return data, where
    #]


def _remove_first_date_change(
    change_data: _np.ndarray | None,
    change_where: list[int] | None,
) -> tuple[_np.ndarray, list[int], ]:
    #[
    if change_where and 0 in change_where:
        where = change_where.index(0)
        change_where.pop(where)
        change_data = _np.delete(change_data, where, axis=0)
        if not change_where:
            change_data = None
            change_where = None
    return change_data, change_where
    #]




"""
Time series mixin for moving sum, average, product
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Callable, )
import numpy as _np
import functools as _ft

from .. import dates as _dates
from . import _functionalize
#]


__all__ = ()


class MovingMixin:
    """
    Time series mixin for moving sum, average, product
    """
    #[

    def mov_sum(self, window: int | None = None, ) -> Self:
        """
        Moving sum of time current and lagged series observations
        """
        return self._moving(_np.sum, window=window, )

    def mov_avg(self, window: int | None = None, ) -> Self:
        """
        Moving average of time current and lagged series observations
        """
        return self._moving(_np.mean, window=window, )

    def mov_mean(self, window: int | None = None, ) -> Self:
        """
        Moving average of time current and lagged series observations
        """
        return self._moving(_np.mean, window=window, )

    def mov_prod(self, window: int | None = None, ) -> Self:
        """
        Moving product of time current and lagged series observations
        """
        return self._moving(_np.prod, window=window, )

    def _moving(
        self,
        func: Callable,
        /,
        window: int | None = None,
    ) -> Self:
        """
        Backend function for moving sum, average, product
        """
        window = (
            window
            if window is not None
            else self._get_default_moving_window()
        )
        window_length = -window
        data = _np.pad(
            self.data,
            pad_width=((window_length-1, 0), (0, 0)),
            mode="constant",
            constant_values=_np.nan,
        )
        data_windows =_np.lib.stride_tricks.sliding_window_view(
            data,
            window_shape=window_length,
            axis=0,
        )
        new_data = func(data_windows, axis=2, )
        self._replace_data(new_data, )

    def _get_default_moving_window(self, /, ) -> int:
        """
        Derive default moving window from time series frequency
        """
        return (
            -self.frequency.value
            if self.frequency is not None and self.frequency.value > 0
            else -4
        )

    #]


for n in ("mov_sum", "mov_avg", "mov_mean", "mov_prod", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )




"""
"""


#[
from __future__ import annotations

import numpy as _np
import dataclasses as _dc
#]


__all__ = (
    "DimensionNames",
)


@_dc.dataclass
class DimensionNames:
    """
    """
    #[
    rows: tuple[str, ...] | None = None
    columns: tuple[str, ...] | None = None

    def select(
        self,
        array: _np.ndarray,
        select_names: tuple[tuple[str, ...], tuple[str, ...]] | tuple[str, ...],
    ) -> _np.ndarray:
        """
        """
        select_row_names, select_column_names = _extract_row_and_column_names(select_names, )
        row_index = _lookup_names(self.rows, select_row_names, )
        column_index = _lookup_names(self.columns, select_column_names, )
        select_array = _np.copy(array)
        select_array = select_array[row_index, :, ...]
        select_array = select_array[:, column_index, ...]
        return select_array
    #]


def _extract_row_and_column_names(
    names: tuple[tuple[str, ...], tuple[str, ...]] | tuple[str, ...],
    /,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    """
    """
    if isinstance(names, str):
        row_names = (names, )
        column_names = (names, )
    elif isinstance(names[0], str):
        row_names = names
        column_names = names
    else:
        row_names = names[0]
        column_names = names[1]
    return row_names, column_names


def _lookup_names(
    names: tuple[str, ...],
    select_names: tuple[str, ...],
) -> tuple[int, ...]:
    """
    """
    return tuple(
        names.index(n)
        for n in select_names
    )




"""
Merge mixin
"""


#[
from __future__ import annotations

from typing import (Self, Iterable, Literal, )
from .. import wrongdoings as _wrongdoings
from ..series import main as _series
from . import main as _databoxes
#]


class MergeMixin:
    """
    """
    #[

    def merge(
        self: Self,
        they: Iterable[Self] | Self,
        action: Literal["hstack", "replace", "discard", "silent", "warning", "error", "critical", ] = "hstack",
        **kwargs,
    ) -> None:
        """
        """
        action_func = _MERGE_ACTIONS[action]
        stream = _wrongdoings.create_stream(action, when_no_stream="silent", )(
            "Duplicate keys when merging databoxes",
        )
        if hasattr(they, "items", ):
            they = (they, )
        for t in they:
            for key, value in t.items():
                if key in self:
                    action_func(self, key, value, stream, **kwargs, )
                else:
                    self[key] = value
        stream._raise()

    #]


def _merge_hstack(
    self,
    key: str,
    value: Any,
    /,
    *args,
) -> None:
    """
    """
    #[
    if isinstance(value, _series.Series, ):
        self[key] = self[key] | value
        return
    if not isinstance(self[key], list):
        self[key] = [self[key], ]
    if not isinstance(value, list):
        value = [value, ]
    self[key] += value
    #]


def _merge_replace(
    self,
    key: str,
    value: Any,
    /,
    *args,
) -> None:
    """
    """
    #[
    self[key] = value
    #]


def _merge_discard(
    self,
    key: str,
    value: Any,
    /,
    *args,
) -> None:
    """
    """
    #[
    pass
    #]


def _merge_report(
    self,
    key: str,
    value: Any,
    stream,
    /,
    *args,
) -> None:
    """
    """
    #[
    stream.add(key, )
    #]


_MERGE_ACTIONS = {
    "hstack": _merge_hstack,
    "replace": _merge_replace,
    "discard": _merge_discard,
    "silent": _merge_report,
    "warning": _merge_report,
    "error": _merge_report,
    "critical": _merge_report,
}




"""
Documentation introspection tools
"""


#[
from __future__ import annotations

from collections.abc import (Callable, )
import re as _re
import os as _os
#]


_TAGLINE_PATTERN = _re.compile(r"==(.*?)==", )


def reference(**kwargs, ) -> Callable:
    def _decorate(callable_object, ) -> Callable:
        callable_object._pages_reference = True
        for k, v in kwargs.items():
            setattr(callable_object, f"_pages_{k}", v)
        if not hasattr(callable_object, "_pages_call_name", ):
            callable_object._pages_call_name = callable_object.__name__
        if not hasattr(callable_object, "_pages_category", ):
            callable_object._pages_category = None
        callable_object._pages_tagline = _extract_tagline(callable_object, )

        return callable_object
    return _decorate


def _extract_tagline(callable_object, ) -> str:
    m = _TAGLINE_PATTERN.search(callable_object.__doc__, )
    return m.group(1, ) if m else ""




"""
Data arrays with row and column names
"""


#[
from __future__ import annotations

from typing import (Self, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, Iterator, )
import numpy as _np
import dataclasses as _dc
import numpy as _np
import functools as _ft
import itertools as _it

from ..series import main as _series
from ..databoxes import main as _databoxes
from ..plans import main as _plans
from ..incidences import main as _incidences
from ..conveniences import iterators as _iterators
from .. import dates as _dates
from .. import has_variants as _has_variants

from . import _invariants as _invariants
from . import _variants as _variants
#]


__all__ = (
    "Dataslate",
)


class SlatableProtocol(Protocol, ):
    """
    """
    def get_min_max_shift(self, /, ) -> tuple[int, int]: ...
    def get_databox_names(self, /, ) -> tuple[str, ...]: ...
    def get_fallbacks(self, /, ) -> dict[str, Number]: ...
    def get_overwrites(self, /, ) -> dict[str, Number]: ...
    def get_scalar_names(self, /, ) -> Iterable[str, ...]: ...
    def create_qid_to_logly(self, /, ) -> dict[int, bool]: ...


@_dc.dataclass
class Dataslate(
    _has_variants.HasVariantsMixin,
):
    """
    """
    #[

    _invariant: _invariants.Invariant | None = None
    _variants: list[_variants.Variant] | None = None

    @classmethod
    def skeleton(
        klass,
        other: Self,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = other._invariant
        self._variants = []
        return self

    @classmethod
    def from_databox(
        klass,
        databox: _databoxes.Databox | dict,
        names: Iterable[str],
        dates: Iterable[_dates.Dater],
        /,
        num_variants: int = 1,
        fallbacks: dict[str, Number] | None = None,
        overwrites: dict[str, Number] | None = None,
        **kwargs,
    ) -> Iterator[Self]:
        """
        """
        names = tuple(names or databox.keys())
        dates = tuple(dates)
        #
        from_to = dates[0], dates[-1] if dates else ()
        item_iterator = \
            _ft.partial(_slate_value_variant_iterator, from_to=from_to, )
        #
        databox_variant_iterator = \
            _databoxes.Databox.iter_variants(databox, item_iterator=item_iterator, names=names, )
        #
        fallbacks_variant_iterator = (
            _databoxes.Databox.iter_variants(fallbacks, item_iterator=item_iterator, names=names, )
            if fallbacks else _it.repeat(None, )
        )
        #
        overwrites_variant_iterator = (
            _databoxes.Databox.iter_variants(overwrites, item_iterator=item_iterator, names=names, )
            if overwrites else _it.repeat(None, )
        )
        #
        zipped = zip(
            range(num_variants),
            databox_variant_iterator,
            fallbacks_variant_iterator,
            overwrites_variant_iterator,
        )
        #
        self = klass()
        self._invariant = _invariants.Invariant(databox, names, dates, **kwargs, )
        self._variants = []
        for vid, databox_v, fallbacks_v, overwrites_v in zipped:
            new_variant = _variants.Variant(
                databox_v, self._invariant,
                fallbacks=fallbacks_v,
                overwrites=overwrites_v,
            )
            self._variants.append(new_variant, )
        return self

    @classmethod
    def from_databox_for_slatable(
        klass,
        slatable: SlatableProtocol,
        databox: _databoxes.Databox | dict,
        base_span: Iterable[_dates.Dater],
        /,
        extra_databox_names: Iterable[str] | None = None,
        **kwargs,
    ) -> Iterator[Self]:
        """
        """
        names = tuple(slatable.get_databox_names())
        if extra_databox_names:
            names = names + tuple(i for i in extra_databox_names if i not in names)
        dates, base_columns, *min_max_shift = _get_extended_range(slatable, base_span, )
        qid_to_logly = slatable.create_qid_to_logly()
        #
        return klass.from_databox(
            databox, names, dates,
            fallbacks=slatable.get_fallbacks(),
            overwrites=slatable.get_overwrites(),
            base_columns=base_columns,
            scalar_names=slatable.get_scalar_names(),
            min_max_shift=min_max_shift,
            qid_to_logly=qid_to_logly,
            **kwargs,
        )

    @property
    def missing_names(self, /, ) -> tuple[str, ...] | None:
        return self._invariant.missing_names

    @property
    def descriptions(self, /, ) -> tuple[str, ...] | None:
        return self._invariant.descriptions

    @property
    def boolex_logly(self, /, ) -> tuple[str, ...] | None:
        return self._invariant.boolex_logly

    @property
    def dates(self, /, ) -> tuple[str]:
        """
        """
        return self._invariant.dates

    @property
    def names(self, /, ) -> tuple[str]:
        """
        """
        return self._invariant.names

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self.names)

    @property
    def num_initials(self, /, ) -> int:
        """
        """
        return -self._invariant.min_max_shift[0]

    @property
    def num_terminals(self, /, ) -> int:
        """
        """
        return self._invariant.min_max_shift[1]

    def remove_initial_data(self, /, ) -> None:
        """
        """
        if self.num_initials:
            self.data = self.data[:, self.num_initials:]

    def to_databox(
        self,
        /,
        target_databox: _databoxes.Databox | None = None,
    ) -> _databoxes.Databox:
        """
        Add data from a dataslate to a new or existing databox
        """
        #[
        if target_databox is None:
            target_databox = _databoxes.Databox()
        num_names = self.num_names
        num_variants = self.num_variants
        start_date = self._invariant.dates[0]
        descriptions = self._invariant.descriptions
        zipped = enumerate(zip(self._invariant.names, self._invariant.databox_value_creator, ), )
        for record_id, (name, creator) in zipped:
            values = _np.vstack(tuple(self._generate_record_from_all_variants(record_id, )))
            target_databox[name] = creator(
                num_variants=num_variants,
                start_date=start_date,
                values=values,
                description=descriptions[record_id],
            )
        return target_databox
        #]

    def _generate_record_from_all_variants(
        self,
        record_id: int,
        /,
    ) -> Iterator[_np.ndarray]:
        """
        """
        return ( v.retrieve_record(record_id, ) for v in self._variants )

    for n in ["num_periods", "from_to", "num_row", "base_slice", "base_columns"]:
        exec(f"@property\ndef {n}(self, /, ): return self._invariant.{n}", )

    def get_data_variant(
        self,
        vid: int,
    ) -> _np.ndarray:
        """
        """
        return self._variants[vid].data

    # @property
    # def num_periods(self, /, ) -> int:
    #     """
    #     """
    #     return self._invariant.num_periods

    # @property
    # def from_to(self, /, ) -> tuple[_dates.Dater, _dates.Dater]:
    #     """
    #     """
    #     return self._invariant.from_to

    # @property
    # def num_rows(self, /, ) -> int:
    #     """
    #     """
    #     return self._invariant.num_names

    # @property
    # def base_slice(self, /, ) -> slice:
    #     """
    #     """
    #     return self._invariant.base_slice

    # @property
    # def base_columns(self, /, ) -> tuple[int, ...]:
    #     """
    #     """
    #     return self._invariant.base_columns

    def remove_initial(self, /, ) -> None:
        """
        """
        remove = -self._invariant.min_max_shift[0]
        self.remove_periods_from_start(remove, )

    def remove_terminal(self, /, ) -> None:
        """
        """
        remove = self._invariant.min_max_shift[1]
        self.remove_periods_from_end(remove, )

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return self._invariant.create_name_to_row()

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove < 0:
            raise ValueError("Cannot remove negative number of columns")
        self._invariant.remove_periods_from_start(remove, )
        for v in self._variants:
            v.remove_periods_from_start(remove, )

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove < 0:
            raise ValueError("Cannot remove negative number of columns")
        self._invariant.remove_periods_from_end(remove, )
        for v in self._variants:
            v.remove_periods_from_end(remove, )

    #]


#class VerticalDataslate(_Dataslate, ):
#    """
#    """
#    #[
#    _record_reshape = (-1, 1, )
#
#    @property
#    def row_dates(self, /, ) -> tuple[_dates.Dater, ...]:
#        """
#        """
#        return self.dates
#
#    @property
#    def column_names(self, /, ) -> tuple[str]:
#        """
#        """
#        return self.names
#
#    @property
#    def base_rows(self, /, ) -> tuple[int, ...]:
#        """
#        """
#        return self._base_indexes
#
#    def retrieve_record(
#        self,
#        record_id: int,
#        /,
#        slice_: slice | None = None,
#    ) -> _np.ndarray:
#        """
#        """
#        if slice_ is None:
#            return self.data[:, record_id]
#        else:
#            return self.data[slice_, record_id]
#
#    def _remove_terminal_data(
#        self,
#        last_base_index: int,
#        /,
#    ) -> None:
#        self.data = self.data[:last_base_index+1, :]
#
#    def store_record(
#        self,
#        values: _np.ndarray,
#        record_id: int,
#        /,
#        slice_: slice | None = None,
#    ) -> None:
#        """
#        """
#        if slice_ is None:
#            self.data[:, record_id] = values
#        else:
#            self.data[slice_, record_id] = values
#
#    #]

def _get_extended_range(
    slatable: SlatableProtocol,
    base_span: Iterable[_dates.Dater],
    /,
) -> tuple[Iterable[_dates.Dater], tuple[int, ...]]:
    """
    """
    base_span = tuple(t for t in base_span)
    num_base_periods = len(base_span)
    min_shift, max_shift = slatable.get_min_max_shifts()
    if min_shift == 0:
        min_shift = -1
    min_base_date = min(base_span)
    max_base_date = max(base_span)
    start_date = min_base_date + min_shift
    end_date = max_base_date + max_shift
    base_columns = tuple(_dates.date_index(base_span, start_date))
    extended_dates = tuple(_dates.Ranger(start_date, end_date))
    return extended_dates, base_columns, min_shift, max_shift


def _slate_value_variant_iterator(
    value: Any,
    /,
    from_to: tuple[_dates.Dater, _dates.Dater],
) -> Iterator[Any]:
    """
    """
    #[
    if hasattr(value, "iter_data_variants_from_to"):
        return value.iter_data_variants_from_to(from_to, )
    elif isinstance(value, Iterable):
        return _iterators.exhaust_then_last(value, )
    else:
        return _iterators.exhaust_then_last([], value, )
    #]


def _retrieve_descriptions(
    databox: _databoxes.Databox | dict,
    names: Iterable[str],
) -> tuple[str]:
    """
    """
    #[
    def _retrieve_item_description(n: str, /, ) -> str:
        if n in databox and hasattr(databox[n], "get_description"):
            return databox[n].get_description()
        else:
            return None
    #
    return { n: _retrieve_item_description(n, ) for n in names }
    #]


def retrieve_vector_from_data_array(
    data: _np.ndarray,
    tokens: tuple[str, ...],
    index_zero: int,
    /,
) -> _np.ndarray:
    """
    """
    rows, columns = _incidences.rows_and_columns_from_tokens(tokens, index_zero, )
    return data[rows, columns]


def store_vector_in_data_array(
    vector: _np.ndarray,
    data: _np.ndarray,
    tokens: Iterable[_incidences.Incidence],
    index_zero: int,
) -> None:
    """
    """
    rows, columns = _incidences.rows_and_columns_from_tokens(tokens, index_zero, )
    data[rows, columns] = vector




"""
Plotly interface to time series objects
"""


#[
from __future__ import annotations

from typing import (Self, Sequence, Iterable, Any, )
from types import EllipsisType
import os as _os
import json as _js
import copy as _cp
import plotly.graph_objects as _pg
import itertools as _it

from .. import dates as _dates
from .. import plotly_wrap as _plotly_wrap
#]


_PLOTLY_STYLES_FOLDER = _os.path.join(_os.path.dirname(__file__), "plotly_styles")
_PLOTLY_STYLES = {
    "layouts": {},
}
with open(_os.path.join(_PLOTLY_STYLES_FOLDER, "plain_layout.json", ), "rt", ) as fid:
    _PLOTLY_STYLES["layouts"]["plain"] = _js.load(fid, )


def _line_plot(color: str, **settings) -> _pg.Scatter:
    """
    """
    settings = {"mode": "lines+markers", } | settings
    return _pg.Scatter(line_color=color, **settings, )


def _bar_plot(color: str, **settings) -> _pg.Bar:
    """
    """
    return _pg.Bar(marker_color=color, **settings, )


_PLOTLY_TRACES_FACTORY = {
    "line": _line_plot,
    "bar": _bar_plot,
}


#_COLOR_ORDER = [
#    "rgb(  0,   114,   189)",
#    "rgb(217,    83,    25)",
#    "rgb(237,   177,    32)",
#    "rgb(126,    47,   142)",
#    "rgb(119,   172,    48)",
#    "rgb( 77,   190,   238)",
#    "rgb(162,    20,    47)",
#]


_COLOR_ORDER = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]


class Mixin:
    """
    """
    #[

    def plot(
        self,
        *,
        span: Iterable[_dates.Dater] | EllipsisType = ...,
        figure: _pg.Figure | None = None,
        figure_title: str | None = None,
        subplot_title: str | None = None,
        legend: Iterable[str] | None = None,
        show_figure: bool = True,
        show_legend: bool | None = None,
        subplot: tuple[int, int] | int | None = None,
        xline = None,
        type: Literal["line", "bar"] = "line",
        traces: tuple(dict[str, Any], ) | None = None,
        freeze_span: bool = False,
        reverse_plot_order: bool = False,
    ) -> _pg.Figure:
        """
        """
        span = self._resolve_dates(span, )
        span = [ t for t in span ]
        frequency = span[0].frequency
        num_variants = self.num_variants
        data = self.get_data(span, )
        date_strings = [ t.to_plotly_date() for t in span ]
        date_format = span[0].frequency.plotly_format
        figure = _pg.Figure() if figure is None else figure
        tile, index = _plotly_wrap.resolve_subplot(figure, subplot, )
        row, column = (tile[0]+1, tile[1]+1, ) if tile is not None else (None, None, )

        if show_legend is None:
            show_legend = legend is not None

        traces = (traces, ) if isinstance(traces, dict) else traces
        color_cycle = _it.cycle(_COLOR_ORDER)
        traces_cycle = _it.cycle(traces or ({}, ))

        loop = zip(range(num_variants, ), color_cycle, traces_cycle, )
        if reverse_plot_order:
            loop = reversed(list(loop))

        for i, color, ts in loop:
            traces_settings = {
                "x": date_strings,
                "y": data[:, i],
                "name": legend[i] if legend else None,
                "showlegend": show_legend,
                "xhoverformat": date_format,
            }
            traces_settings |= ts or {}
            traces_object = _PLOTLY_TRACES_FACTORY[type](color, **traces_settings, )
            figure.add_trace(traces_object, row=row, col=column, )

        # REFACTOR
        xaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["xaxis"])
        yaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["yaxis"])
        layout = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"])
        del layout["xaxis"]
        del layout["yaxis"]

        xaxis["tickformat"] = date_format
        xaxis["ticklabelmode"] = "period"
        if freeze_span:
            xaxis["range"] = [date_strings[0], date_strings[-1], ]
            xaxis["autorange"] = False

        figure.update_xaxes(xaxis, row=row, col=column, )
        figure.update_yaxes(yaxis, row=row, col=column, )
        if figure_title is not None:
            figure.update_layout(title={"text": figure_title}, )
        figure.update_layout(layout or {}, )

        if subplot_title:
            _update_subplot_title(figure, subplot_title, index, )

        if xline:
            xline = (
                xline.to_plotly_date()
                if hasattr(xline, "to_plotly_date")
                else xline
            )
            figure.add_vline(xline, )

        if show_figure:
            figure.show()

        return figure

    #]


def _update_subplot_title(
    figure: _pg.Figure,
    subplot_title: str,
    index: int,
    /,
) -> None:
    """
    """
    annotation = next(figure.select_annotations(index, ), None, )
    if annotation:
        annotation.text = subplot_title




"""
"""


#[
from __future__ import annotations

import dataclasses as _dc
from typing import (Callable, )
from numbers import (Real, )

from .. import dates as _dates
from .. import has_variants as _has_variants
from ..conveniences import descriptions as _descriptions
from ..series import main as _series
from ..databoxes import main as _databoxes
#]


@_dc.dataclass(slots=True, )
class Invariant:
    """
    """
    #[

    missing_names: tuple[str, ...] | None = None
    descriptions: Iterable[str] | None = None
    boolex_logly: tuple[bool, ...] | None = None
    index_base_columns: tuple[int, ...] | None = None
    names: tuple[str] | None = None
    dates: tuple[_dates.Dater, ...] | None = None
    min_max_shift: tuple[int, int] | None = None
    databox_value_creator: tuple[Callable, ...] | None = None

    def __init__(
        self,
        databox: _databoxes.Databox,
        names: Iterable[str],
        dates: Iterable[_dates.Dater],
        /,
        *,
        base_columns: Iterable[int] = None,
        descriptions: dict[str, str | None] | None = None,
        scalar_names: Iterable[str] | None = None,
        qid_to_logly: dict[int, bool | None] | None = None,
        min_max_shift: tuple[int, int] = (0, 0),
    ) -> None:
        """
        """
        self.names = tuple(names)
        self.dates = dates
        self.missing_names = tuple(databox.get_missing_names(self.names, ))
        self.index_base_columns = tuple(i in base_columns for i in range(self.num_periods))
        self._populate_descriptions(descriptions, )
        self._populate_boolex_logly(qid_to_logly, )
        self.min_max_shift = min_max_shift
        self._populate_databox_value_creator(scalar_names, )

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self.dates)

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self.names)

    @property
    def from_to(self, /, ) -> tuple[_dates.Dater, _dates.Dater]:
        """
        """
        return self.dates[0], self.dates[-1]

    @property
    def base_columns(self, /, ) -> tuple[int]:
        """
        """
        return tuple(i for i in range(self.num_periods) if self.index_base_columns[i])

    @property
    def base_slice(self, /, ) -> slice:
        """
        """
        base_columns = self.base_columns
        return slice(base_columns[0], base_columns[-1]+1)

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return {
            name: row
            for row, name in enumerate(self.names, )
        }

    def _populate_descriptions(
        self,
        descriptions: dict[str, str] | None,
    ) -> None:
        """
        """
        descriptions = descriptions or {}
        self.descriptions = tuple(
            descriptions.get(n, "") or ""
            for n in self.names
        )

    def _populate_boolex_logly(
        self,
        qid_to_logly: dict[int, bool] | None,
        /,
    ) -> None:
        """
        """
        qid_to_logly = qid_to_logly or {}
        self.boolex_logly = tuple(
            qid_to_logly.get(i, False) or False
            for i in range(len(self.names), )
        )

    def _populate_databox_value_creator(
        self,
        scalar_names: Iterable[str] | None,
        /,
    ) -> None:
        """
        """
        if not scalar_names:
            scalar_names = ()
        self.databox_value_creator = tuple(
            _scalar_from_values if n in scalar_names else _series_from_values
            for n in self.names
        )

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.dates = self.dates[remove:]
            self.index_base_columns = self.index_base_columns[remove:]

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.dates = self.dates[:-remove]
            self.index_base_columns = self.index_base_columns[:-remove]


def _series_from_values(
    num_variants: int,
    values: _np.ndarray,
    start_date: _dates.Dater,
    description: str,
) -> _series.Series:
    """
    """
    return _series.Series(
        num_variants=num_variants,
        start_date=start_date,
        values=values.T,
        description=description,
    )


def _scalar_from_values(
    num_variants: int,
    values: _np.ndarray,
    **kwargs,
) -> list[Real] | Real:
    """
    """
    scalars = list(values[:, 0])
    is_singleton = _has_variants.is_singleton(num_variants, )
    return _has_variants.unwrap_singleton(scalars, is_singleton, )




"""
"""


#[
from __future__ import annotations

from typing import (Callable, )
from numbers import (Real, )
import numpy as _np

from .. import dates as _dates
from . import main as _series
from . import _functionalize
#]


__all__ = ()


class DiffMixin:
    """
    """
    #[

    def _shifted_op(
        self,
        shift: int | str,
        func: Callable,
        /,
    ) -> None:
        """
        """
        _catch_invalid_shift(shift, )
        other = self.copy()
        other.shift(shift, )
        self._binop(other, func, new=self, )

    def diff(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        """
        """
        self._shifted_op(shift, lambda x, y: x - y, )

    def diff_log(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        """
        """
        self._shifted_op(shift, lambda x, y: _np.log(x) - _np.log(y), )

    def roc(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        """
        """
        self._shifted_op(shift, lambda x, y: x/y, )

    def pct(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        """
        """
        self._shifted_op(shift, lambda x, y: 100*(x/y - 1), )

    #]


for n in ("diff", "diff_log", "pct", "roc", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


_CUMULATIVE_FACTORY = {
    "diff": {
        "forward": lambda x_past, change_curr: x_past + change_curr,
        "backward": lambda x_future, change_future: x_future - change_future,
        "initial": 0,
    },
    "diff_log": {
        "forward": lambda x_past, change_curr: x_past * exp(change_curr),
        "backward": lambda x_future, change_future: x_future / exp(change_future),
        "initial": 0,
    },
    "pct": {
        "forward": lambda x_past, change_curr: x_past * (1 + change_curr/100),
        "backward": lambda x_future, change_future: x_future / (1 + change_future/100),
        "initial": 1,
    },
    "roc": {
        "forward": lambda x_past, change_curr: x_past * change_curr,
        "backward": lambda x_future, change_future: x_future / change_future,
        "initial": 1,
    },
}


class CumMixin:
    """
    """
    #[

    def cum_diff(self, *args, **kwargs, ) -> None:
        """
        """
        self._cumulate("diff", *args, **kwargs, )

    def cum_diff_log(self, *args, **kwargs, ) -> None:
        """
        """
        self._cumulate("diff_log", *args, **kwargs, )

    def cum_pct(self, *args, **kwargs, ) -> None:
        """
        """
        self._cumulate("pct", *args, **kwargs, )

    def cum_roc(self, *args, **kwargs, ) -> None:
        """
        """
        self._cumulate("roc", *args, **kwargs, )

    def _cumulate(
        self,
        func_name: str,
        shift: int | str = -1,
        initial: Real | _series.Series | None = None,
        range: _dates.Ranger | None = None,
    ) -> None:
        """
        """
        _catch_invalid_shift(shift, )
        range = _dates.Ranger(None, None, ) if range is None else range
        range = range.resolve(self, )
        direction = range.direction
        factory = _CUMULATIVE_FACTORY[func_name]
        cum_func = factory[direction]
        initial = factory["initial"] if initial is None else initial
        if direction == "forward":
            self._cumulate_forward(shift, cum_func, initial, range, )
        elif direction == "backward":
            self._cumulate_backward(shift, cum_func, initial, range, )

    def _cumulate_forward(self, shift, cum_func, initial, range, /, ) -> None:
        """
        """
        shifted_range = tuple(t.shift(shift, ) for t in range)
        initial_range = _dates.Ranger(min(shifted_range), range.end_date, )
        orig = self.copy()
        self.empty()
        self.set_data(initial_range, initial)
        for t, sh in zip(range, shifted_range):
            new_data = cum_func(self.get_data(sh, ), orig.get_data(t, ), )
            self.set_data(t, new_data)

    def _cumulate_backward(self, shift, cum_func, initial, shifted_backward_range, /, ) -> None:
        """
        """
        orig_range_shifted = _dates.Ranger(self.start_date, self.end_date, -1, )
        orig_range_shifted.shift(shift, )
        shifted_backward_range = shifted_backward_range.resolve(orig_range_shifted, )
        backward_range = shifted_backward_range.copy()
        backward_range.shift(-shift, )
        initial_range = _dates.Ranger(min(shifted_backward_range), backward_range.start_date, )
        orig = self.copy()
        self.empty()
        self.set_data(initial_range, initial, )
        for t, sh in zip(backward_range, shifted_backward_range, ):
            new_data = cum_func(self.get_data(t, ), orig.get_data(t, ), )
            self.set_data(sh, new_data, )


for n in ("cum_diff", "cum_diff_log", "cum_pct", "cum_roc", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _catch_invalid_shift(shift: int | str, ):
    """
    """
    if not isinstance(shift, str) and (int(shift) != shift or shift >= 0):
        raise ValueError("Time shift must be a negative integer or a string")




"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, )

from .. import equations as _equations
from .. import quantities as _quantities
from ..explanatories import main as _explanatories
from ..conveniences import descriptions as _descriptions
#]


class Invariant(
    _descriptions.DescriptionMixin,
):
    """
    """
    #[

    __slots__ = (
        "explanatories",
        "all_names",
        "_context",
        "_description",
    )

    def __init__(self, /, ) -> None:
        """
        """
        self.explanatories = ()
        self.all_names = ()
        self._context = {}
        self._description = ""

    @classmethod
    def from_equations(
        klass,
        equations: Iterable[_equations.Equation],
        quantities: Iterable[_quantities.Quantity] | None,
        /,
        context: dict[str, Any] | None = None,
        description: str | None = None,
        **kwargs,
    ) -> None:
        """
        """
        self = klass()
        self._context = context or {}
        self.set_description(description, )
        self.explanatories = tuple(
            _explanatories.Explanatory(e, context=self._context, **kwargs, )
            for e in equations
        )
        self.collect_all_names()
        self.finalize_explanatories()
        quantity_names = _quantities.generate_all_quantity_names(quantities, )
        self.parameter_names = tuple(n for n in quantity_names if n in self.all_names)
        return self

    @property
    def num_equations(self, /, ) -> int:
        """
        Number of equations.
        """
        return len(self.explanatories)

    @property
    def lhs_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance.
        """
        return tuple( x.lhs_name for x in self.explanatories )

    @property
    def residual_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance
        """
        return tuple(
            x.residual_name
            for x in self.explanatories
            if x.residual_name is not None
        )

    def collect_all_names(
        self,
        /,
    ) -> None:
        """
        """
        all_names = set()
        for x in self.explanatories:
            all_names.update(x.all_names)
        rhs_only_names = tuple(all_names.difference(self.lhs_names + self.residual_names, ))
        self.all_names = self.lhs_names + rhs_only_names + self.residual_names

    def finalize_explanatories(
        self,
        /,
    ) -> None:
        """
        """
        name_to_qid = self.create_name_to_qid()
        for x in self.explanatories:
            x.finalize(name_to_qid, )

    def reorder_equations(
        self,
        order = Iterable[int],
        /,
    ) -> None:
        self.explanatories = [
            self.explanatories[i]
            for i in order
        ]
        self.collect_all_names()
        self.finalize_explanatories()

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        """
        """
        return { name: i for i, name in enumerate(self.all_names) }

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        """
        """
        return { name: i for i, name in enumerate(self.all_names) }

    #]

